# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from datetime import datetime
from ConfigParser import SafeConfigParser

from storm.locals import create_database, Store

from offspring.tests.helpers import CaptureLoggingTestCase

from offspring.slave import LexbuilderSlave as Slave
from offspring.master.models import (
    User, Project, BuildRequest, ProjectNotificationSubscription)


class StubMailer(object):

    def __init__(self, error=None):
        self.error = error
        self.mails = []

    def mail(self, sender, rcpt, msg):
        if self.error:
            raise self.error
        self.mails.append((sender, rcpt, msg))


class FakeBuilder(object):
    def __init__(self, current_state, build_result=None):
        self.name = u"test-builder"
        self.old_state = self.current_state = current_state
        self.new_state = Slave.STATE_IDLE
        self.scanned = False
        self.machine_type = u"x86_64"
        self.build_result = self.current_job = build_result

    def scan(self):
        self.scanned = True
        return self.old_state, self.new_state

    def getBuildResult(self):
        return self.build_result


class BaseOffspringMasterTestCase(CaptureLoggingTestCase):

    database_uri = (u"postgres://offspring_test:temp1234@localhost/"
                    "offspring_test")

    def setUp(self):
        super(BaseOffspringMasterTestCase, self).setUp()
        self.setup_config()
        self.database = create_database(self.database_uri)
        self.connection = self.database.connect()
        self.drop_tables()
        self.create_tables()
        self.create_db_store()
        self.create_fixtures()

    def setup_config(self):
        """
        Create a base configuration for use in tests.
        """
        self.config = SafeConfigParser()
        self.config.set("DEFAULT", "instance_email", "testing@example.com")
        self.config.set("DEFAULT", "notification_email",
                        "testing@example.com")
        self.config.set("DEFAULT", "instance_name", "test-builder")
        self.config.set("DEFAULT", "discussion_email", "discuss@example.com")
        self.config.set("DEFAULT", "admin_email", "admin@example.com")
        self.config.add_section("web")
        self.config.set("web", "build_results_uri",
                        "https://offspring-share/builds")
        self.config.set("web", "ibs_uri", "http://offspring-web")
        self.config.add_section("master")
        self.log_dir = self.makeDir()
        self.config.set("master", "log_dir", self.log_dir)
        self.config.set("master", "db", self.database_uri)
        self.config.set("master", "logfile", "%s/offspring-master.log"
                        % self.log_dir)
        return self.config

    def create_fixtures(self):
        """
        Create a basic set of fixtures.
        """
        self.user = User()
        self.db_store.add(self.user)
        self.user.email = u"testing@example.com"
        self.user.date_joined = datetime.now()
        self.project = Project(u"test-project")
        self.project.config_url = u"http://example.com/"
        self.project.series = u"test-series"
        self.project.suite = u"test-suite"
        self.project.arch = u"i386"
        self.db_store.add(self.project)

    def tearDown(self):
        self.drop_db_store()
        self.connection.close()
        super(BaseOffspringMasterTestCase, self).tearDown()

    def create_tables(self):
        self.connection.execute(
            "CREATE TABLE auth_user ( "
            "id SERIAL NOT NULL PRIMARY KEY, "
            "date_joined timestamp NOT NULL, "
            "username VARCHAR(30) UNIQUE, "
            "first_name VARCHAR(30), "
            "last_name VARCHAR(30), "
            "password VARCHAR(128), "
            "is_active BOOLEAN NOT NULL DEFAULT TRUE, "
            "is_staff BOOLEAN NOT NULL DEFAULT FALSE, "
            "is_superuser BOOLEAN NOT NULL DEFAULT FALSE, "
            "last_login TIMESTAMP NOT NULL DEFAULT (now() AT TIME ZONE "
              "'UTC'), "
            "email VARCHAR(75) NOT NULL)")

        self.connection.execute(
            "CREATE TABLE projects ( "
            "name VARCHAR(200) NOT NULL, "
            "title VARCHAR(30) NOT NULL, "
            "project_group_id  VARCHAR(200), "
            "priority INTEGER NOT NULL, "
            "arch VARCHAR(10) NOT NULL, "
            "config_url VARCHAR(200) NOT NULL, "
            "launchpad_project_id VARCHAR(200), "
            "series VARCHAR(200) NOT NULL, "
            "suite VARCHAR(200) NOT NULL, "
            "status VARCHAR(200), "
            "notes TEXT, "
            "lp_user TEXT, "
            "lp_ssh_key text, "
            "is_active BOOLEAN NOT NULL DEFAULT TRUE)")

        self.connection.execute(
            "CREATE TABLE project_notification_subscriptions ( "
            "user_id INTEGER NOT NULL, "
            "project_name VARCHAR(200) NOT NULL)")

        self.connection.execute(
            "CREATE TABLE lexbuilders ( "
            "id SERIAL NOT NULL PRIMARY KEY, "
            "name VARCHAR(200) NOT NULL, "
            "uri VARCHAR(200) NOT NULL, "
            "current_state VARCHAR(200) DEFAULT 'UNKNOWN', "
            "previous_state VARCHAR(200) DEFAULT 'UNKNOWN', "
            "machine_type VARCHAR(200), "
            "notes TEXT, "
            "created_at TIMESTAMP NOT NULL DEFAULT (now() AT TIME ZONE "
              "'UTC'), "
            "updated_at TIMESTAMP NOT NULL DEFAULT (now() AT TIME ZONE "
              "'UTC'), "
            "current_job_id INTEGER, "
            "is_okay BOOLEAN NOT NULL, "
            "is_active BOOLEAN NOT NULL DEFAULT TRUE)")

        self.connection.execute(
            "CREATE TABLE buildrequests ("
            "id SERIAL NOT NULL, "
            "project_name VARCHAR(200) NOT NULL, "
            "created_at timestamp with time zone NOT NULL, "
            "requestor_id INTEGER, "
            "reason TEXT, "
            "score INTEGER)")

        self.connection.execute(
            "CREATE TABLE buildresults ("
            "id SERIAL NOT NULL, "
            "name VARCHAR(200), "
            "builder_id INTEGER, "
            "project_name VARCHAR(200) NOT NULL, "
            "started_at TIMESTAMP NOT NULL DEFAULT (now() AT TIME ZONE "
              "'UTC'), "
            "finished_at TIMESTAMP NOT NULL DEFAULT (now() AT TIME ZONE "
              "'UTC'), "
            "requested_at TIMESTAMP NOT NULL DEFAULT (now() AT TIME ZONE "
              "'UTC'), "
            "dispatched_at TIMESTAMP NOT NULL DEFAULT (now() AT TIME ZONE "
              "'UTC'), "
            "requestor_id INTEGER, "
            "reason VARCHAR(200), "
            "result VARCHAR(200), "
            "notes VARCHAR(200), "
            "score INTEGER)")
        self.connection.execute(
            "CREATE TABLE dailybuildorders ("
            "id SERIAL NOT NULL PRIMARY KEY, "
            "name VARCHAR(200) NOT NULL, "
            "hour integer NOT NULL, "
            "CONSTRAINT dailybuildorders_hour_check CHECK ((hour >= 0)))")
        self.connection.execute(
            "CREATE TABLE dailybuildorders_projects ("
            "id SERIAL NOT NULL PRIMARY KEY, "
            "dailybuildorder_id integer NOT NULL, "
            "project_id VARCHAR(200) NOT NULL)")
        self.connection.commit()

    def create_db_store(self):
        self.db_store = Store(self.database)

    def drop_db_store(self):
        self.db_store.rollback()
        self.db_store.close()

    def drop_tables(self):
        for table in ["auth_user", "projects", "lexbuilders",
                      "project_notification_subscriptions",
                      "buildresults", "buildrequests", "dailybuildorders",
                      "dailybuildorders_projects"]:
            try:
                self.connection.execute("DROP TABLE %s" % table)
                self.connection.commit()
            except:
                self.connection.rollback()

    def create_build_request(self, project=None, reason=u"testing"):
        """
        Create a BuildRequest for a given project, with a reason.
        """
        project = project if project else self.project
        build_request = BuildRequest(project, reason=u"testing")
        self.db_store.add(build_request)
        return build_request

    def subscribe_to_project(self, user, project):
        """
        Helper method to subscribe a user to a project.
        """
        subscription = ProjectNotificationSubscription(user, project)
        project.subscriptions.add(subscription)
        return subscription
