# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from ConfigParser import ConfigParser
from datetime import datetime
import os
import platform
import re
import subprocess
import xmlrpclib

from offspring.tests.helpers import CaptureLoggingTestCase

from offspring.slave import (
    LexbuilderSlave, ProjectBuildProcess, LexbuilderSlaveServer,
    ShutdownThread)


class FakeLexbuilderSlaveServer(object):
    def start(self):
        self.started = True

    def stop(self):
        self.stopped = True

    def join(self):
        self.joined = True


class FakeBuild(object):

    class FakeBuildProcess(object):

        def poll(self):
            return None

        def terminate(self):
            self.terminated = True

        def kill(self):
            self.killed = True

    def __init__(self, project_name=None, name=None, started_at=None,
                 finished_at=None, error_code=None,
                 state=ProjectBuildProcess.STATE_SUCCESS):
        self.process = self.FakeBuildProcess()
        self.state = state
        self.startedAt = started_at
        self.finishedAt = finished_at
        self.project_name = project_name
        self.name = name
        self.errorCode = error_code

    def getState(self):
        return self.state

    def getProjectName(self):
        return self.project_name

    def getBuildName(self):
        return self.name

    def tidyUp(self):
        self.tidied = True


class LexbuilderSlaveTestMixin(object):
    """
    Sets up a configuration object usable by Slave tests.
    """

    def setUp(self):
        super(LexbuilderSlaveTestMixin, self).setUp()
        self.log_dir = self.makeDir()
        self.config = ConfigParser()
        self.config.add_section("slave")
        self.config.set("slave", "bin_dir", "/usr/bin")
        self.config.set("slave", "log_dir", self.log_dir)
        self.config.set("slave", "build_command",
                        "%(bin_dir)s/offspring-build")
        self.config.set("slave", "child_wait_timeout", "0")
        self.config.set("slave", "xmlrpc_server_auth", "admin:password")
        self.config.set("slave", "request_timeout", "60")
        self.config.set("slave", "logfile", "%(log_dir)s/offspring-slave-log")
        self.config.set("slave", "port", 8765)


class FakeProjectBuildProcess(object):
    state = ProjectBuildProcess.STATE_ACTIVE


class TestLexbuilderSlave(LexbuilderSlaveTestMixin, CaptureLoggingTestCase):
    """
    Tests for the LexbuilderSlave class.
    """

    def setUp(self):
        super(TestLexbuilderSlave, self).setUp()
        self.fake_server = FakeLexbuilderSlaveServer()
        self.slave = LexbuilderSlave(("localhost", 8760), self.config,
                                     self.fake_server)

    def tearDown(self):
        super(TestLexbuilderSlave, self).tearDown()

    def test_slave_run_starts_server_and_logs(self):
        """
        LexbuilderSlave.run should start the LexbuilderSlaveServer.
        """
        log_file = self.capture_logging()
        self.slave.run()
        self.assertIn("Slave ready to accept builds", log_file.getvalue())
        self.assertTrue(self.fake_server.started)

    def test_slave_stop(self):
        """
        LexbuilderSlave.stop, should terminate any current build, and stop
        the LexbuilderSlaveServer.
        """
        slave_mock = self.mocker.patch(self.slave)
        slave_mock.stopBuild()
        self.mocker.replay()

        self.slave.stop()
        self.assertTrue(self.fake_server.stopped)
        self.assertTrue(self.fake_server.joined)

    def test_slave_stop_build_no_current_build(self):
        """
        LexbuilderSlave.stopBuild should do nothing if there's no current
        build in progress.
        """
        log_file = self.capture_logging()
        self.slave.stopBuild()
        self.assertEqual("", log_file.getvalue())

    def test_slave_stop_build_with_current_build(self):
        """
        LexbuilderSlave.stopBuild should terminate the current build, this
        involves killing the process running the build.
        """
        self.slave.currentBuild = fake_build = FakeBuild()
        log_file = self.capture_logging()

        self.slave.stopBuild()

        self.assertTrue(fake_build.process.terminated)
        self.assertTrue(fake_build.process.killed)
        self.assertTrue(fake_build.tidied)
        self.assertEqual("Killing active build.\n", log_file.getvalue())

    def test_slave_start_build(self):
        """
        LexbuilderSlave.startBuild should start a new build on the server.
        """
        log_file = self.capture_logging()
        build_process_mock = self.mocker.patch(ProjectBuildProcess)
        build_process_mock.start()
        self.mocker.replay()

        fake_url = u"http://example.com/project"
        self.slave.startBuild(u"testing", fake_url)
        self.assertEqual(["testing", fake_url],
                         self.slave.currentBuild.arguments)
        self.assertEqual(
            "Starting build: /usr/bin/offspring-build testing %s\n" % fake_url,
            log_file.getvalue())

    def test_current_build_is_active_no_current_build(self):
        """Test currentBuildIsActive returns False when currentBuild = False"""

        # To start with, currentBuild should be None, so active == False
        self.assertEqual(self.slave.currentBuild, None)
        self.assertEqual(self.slave.currentBuildIsActive(), False)

    def test_current_build_is_active_build_in_active_state(self):
        """Test currentBuildIsActive returns True when build is active"""
        # Set up a fake build process (starts off in STATE_ACTIVE)
        self.slave.currentBuild = FakeProjectBuildProcess()

        # Now we have started a build, currentBuild has been assigned.
        self.assertNotEqual(self.slave.currentBuild, None)
        self.assertEqual(self.slave.currentBuild.state,
                         ProjectBuildProcess.STATE_ACTIVE)
        self.assertEqual(self.slave.currentBuildIsActive(), True)

    def test_current_build_is_active_build_state_not_active(self):
        """Test currentBuildIsActive when currentBuild state != STATE_ACTIVE

        currentBuildIsActive should return False when
        slave.currentBuild.state != STATE_ACTIVE. Clearly, currentBuild needs
        to be valid for this test.

        """

        # Set up a fake build process (starts off in STATE_ACTIVE)
        self.slave.currentBuild = FakeProjectBuildProcess()

        self.slave.currentBuild.state = ProjectBuildProcess.STATE_SUCCESS
        self.assertEqual(self.slave.currentBuildIsActive(), False)

        self.slave.currentBuild.state = ProjectBuildProcess.STATE_FAILED
        self.assertEqual(self.slave.currentBuildIsActive(), False)

    def test_add_user_to_bzr_ssh_url(self):
        """Test lp:project URL form converted into bzr+ssh form adding user."""

        config_url = u"lp:~dooferlad/offspring/config"
        user = "user"

        updated_url = self.slave.addUserToBzrSshUrl(config_url, user)
        expected_url = ("bzr+ssh://user@bazaar.launchpad.net/~dooferlad/" +
                        "offspring/config")
        self.assertEqual(updated_url, expected_url)

    def test_start_build_with_custom_ssh_key(self):
        """Test slave converts lp:project URL into bzr+ssh form adding user.

        When sending a URL to the slave in the form lp:project, and
        a user and SSH key are sent as well, that the slave adds the user
        into the URL so a private config branch can be checked out (URL needs
        to be modified to send the user name through BZR to SSH).
        """
        log_file = self.capture_logging()
        build_process_mock = self.mocker.patch(ProjectBuildProcess)
        build_process_mock.start()

        sshkey_filename = self.makeFile()
        tempfile_mock = self.mocker.replace("tempfile.NamedTemporaryFile")
        tempfile_mock()
        self.mocker.result(open(sshkey_filename, "w"))

        self.mocker.replay()

        config_url = u"lp:~dooferlad/offspring/config"
        args = ["project_name", config_url, "user", "key"]

        self.slave.startBuild(*args)

        new_url = re.sub("lp:",
                          "bzr+ssh://" + args[2] + "@bazaar.launchpad.net/",
                          args[1])
        self.assertEqual(["project_name", new_url, sshkey_filename],
                         self.slave.currentBuild.arguments)
        self.assertEqual(
            "Starting build: /usr/bin/offspring-build project_name %s\n"
            % new_url, log_file.getvalue())

    def test_check_ssh_key_deleted_when_build_finished(self):
        """Test SSH key file lifetime management.

        The SSH key of a project is written to a temporary file when the
        build process starts, this is tested by the first assert. When the
        build has finished or is stopped, the key file is deleted. This is
        tested by the second assert.
        """
        log_file = self.capture_logging()
        build_process_mock = self.mocker.patch(ProjectBuildProcess)
        build_process_mock.start()
        self.mocker.replay()

        config_url = u"lp:~dooferlad/offspring/config"
        args = ["project_name", config_url, "user", "key"]

        self.slave.startBuild(*args)

        file_path = self.slave.currentBuild.ssh_key_file.name
        read_key = open(file_path).read()

        # We should have a key written to a file for the slave to read...
        self.assertEqual(read_key, args[3])

        # Can't use XMLRPC interface here because it checks to see if a build
        # has actually started. We just want to test the clean up logic.
        self.slave.stopBuild()
        # Having stopped the slave, the key should have been erased
        self.assertFalse(os.path.exists(file_path))

    def test_check_ssh_agent(self):
        """Check that SSH agent is started and killed by build.

        In order to pass private keys to SSH, the build script starts
        ssh-agent if a private key file is passed to it. When the build has
        completed, the build script kills SSH agent.

        This test uses the called shell script to test the functions used
        by the build script to start ssh-agent, load the specified key and
        kill the started ssh-agent process.

        """
        # This is a wrapper around a shell script that runs the commands and
        # we just check the output.

        dir_of_this_file = os.path.dirname(os.path.abspath(__file__))
        test_script = os.path.join(dir_of_this_file,
                                   "../../build/tests/test.sh")

        test_script_process = subprocess.Popen(test_script,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.STDOUT)

        stdout = test_script_process.communicate()[0]

        lines = stdout.split("\n")

        # Sample output:
        # Agent pid 9839
        # Identity added: /bla/offspring_test_key (/bla/offspring_test_key)
        # After loading test key, keys are:
        # 2048 <fingerprint> /bla/offspring_test_key (RSA)
        # After killing ssh agent, we shouldn't be able to find any keys...
        # Could not open a connection to your authentication agent.

        # Expect the first line to be of the form "Agent pid [PID]"
        self.assertTrue(re.search("Agent pid \d+", lines[0]),
                        "Test build ssh-agent check failed: Expected to find" +
                        "\nAgent pid <PID>\nfound:\n" + lines[0])

        # Expect the second line to be of the form:
        # Identity added: <path to...>/offspring_test_key
        self.assertTrue(re.search("Identity added:.*?offspring_test_key",
                                  lines[1]),
                        "Test build ssh-agent check failed: Expected to find" +
                        "\nIdentity added: <path to...>/offspring_test_key" +
                        "\nfound:\n" + lines[1])

        # Ignore the third line, it is just a fixed message
        # Expect the fourth line to be:
        # <key fingerprint> /path/to/offspring_test_key (RSA)
        self.assertTrue(re.search("offspring_test_key \(RSA\)", lines[3]),
                        "Test build ssh-agent check failed: Expected to find" +
                        "\noffspring_test_key (RSA)\nfound:\n" + lines[3])

        # Ignore the fifth line, it is just a fixed message
        # Expect the sixth line to be:
        # "Could not open a connection to your authentication agent."
        search_string = (
            "^Could not open a connection to your authentication agent\.$")
        message = ("Test build ssh-agent check failed: Expected to find" +
                   "\nCould not open a connection to your authentication" +
                   "agent\n, found:\n" + lines[5])
        self.assertTrue(re.search(search_string, lines[5]), message)


class FakeLexbuilderSlave(object):
    """Fake slave for use in tests."""
    def __init__(self, state=LexbuilderSlave.STATE_IDLE):
        self.project_name = None
        self.config_url = None
        self.stopped = False
        self.state = state

    def startBuild(self, project_name, config_url):
        self.project_name = project_name
        self.config_url = config_url
        self.started = True

    def stopBuild(self):
        self.stopped = True

    def currentBuildIsActive(self):
        """Return True if current build is active, else False."""
        if self.currentBuild is not None:
            return self.currentBuild.state == ProjectBuildProcess.STATE_ACTIVE
        return False


class TestLexbuilderSlaveServer(LexbuilderSlaveTestMixin,
                                CaptureLoggingTestCase):
    """
    Tests for the LexbuilderSlaveServer class.
    """

    def setUp(self):
        super(TestLexbuilderSlaveServer, self).setUp()
        self.fake_slave = FakeLexbuilderSlave()

    def create_slave_server(self):
        """
        Instantiate a slave server for the fake_slave.
        """
        return LexbuilderSlaveServer(
            self.fake_slave, ("localhost", 8760), self.config)

    def test_create(self):
        """
        We can instantiate a new LexbuilderSlaveServer.
        """
        log_file = self.capture_logging()
        slave_server = self.create_slave_server()
        self.assertEqual(self.fake_slave, slave_server.slave)
        self.assertEqual(("127.0.0.1", 8760), slave_server.server_address)
        self.assertEqual(self.config, slave_server.config)
        self.assertEqual("", log_file.getvalue())

    def test_api_start_build(self):
        """
        If the server doesn't have an active build, then api_startBuild
        should start a build on the slave with the correct project name
        and configuration URL.
        """
        self.fake_slave.currentBuild = FakeBuild()
        slave_server = self.create_slave_server()

        response = slave_server.api_startBuild("testing",
                                               "http://example.com/")
        self.assertEqual(LexbuilderSlaveServer.REQUEST_OKAY, response)
        self.assertEqual("testing", self.fake_slave.project_name)
        self.assertEqual("http://example.com/", self.fake_slave.config_url)
        self.assertTrue(self.fake_slave.started)

    def test_api_start_build_with_busy_server(self):
        """
        If the server has an active build in progress, api_startBuild should
        return an error state.
        """
        slave_server = self.create_slave_server()

        self.fake_slave.currentBuild = FakeBuild(
            state=ProjectBuildProcess.STATE_ACTIVE)

        response = slave_server.api_startBuild("testing",
                                               "http://example.com/")
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR, response)
        self.assertFalse(self.fake_slave.project_name)
        self.assertFalse(self.fake_slave.config_url)

    def test_api_stop_build(self):
        """
        If the SlaveServer has an active build, api_stopBuild should end the
        current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild(
            state=ProjectBuildProcess.STATE_ACTIVE)

        response = slave_server.api_stopBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR, response)
        self.assertFalse(self.fake_slave.stopped)

    def test_api_stop_build_with_correct_platform_version(self):
        """
        If platform.version is > 2.6, and we have a current build, then we
        can stop the build slave.

        FIXME: Add documentation for why?
        """
        version_mock = self.mocker.replace("platform.version")
        version_mock()
        self.mocker.result("2.7")
        self.mocker.replay()
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild(
            state=ProjectBuildProcess.STATE_ACTIVE)

        response = slave_server.api_stopBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_OKAY, response)
        self.assertTrue(self.fake_slave.stopped)

    def test_api_stop_build_no_current_build(self):
        """
        If the SlaveServer does not have an active build, api_stopBuild should
        return an error state.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()

        response = slave_server.api_stopBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR, response)
        self.assertFalse(self.fake_slave.stopped)

    def test_api_get_builder_name(self):
        """
        LexbuilderSlaveServer.api_getBuilderName should return the network name
        for this machine.
        """
        slave_server = self.create_slave_server()
        self.assertEqual(platform.node(), slave_server.api_getBuilderName())

    def test_api_get_machine_type(self):
        """
        LexbuilderSlaveServer.api_getMachineType should return the platform the
        builder is running on.
        """
        slave_server = self.create_slave_server()
        self.assertEqual(platform.machine(), slave_server.api_getMachineType())

    def test_api_get_builder_state(self):
        """
        LexbuilderSlaveServer.api_getBuilderState should return the current
        state of the underlying LexBuilderSlave.
        """
        slave_server = self.create_slave_server()
        self.assertEqual(LexbuilderSlave.STATE_IDLE,
                         slave_server.api_getBuilderState())

    def test_api_get_build_project_name(self):
        """
        LexbuilderSlaveServer.api_getBuildProjectName should return the name of
        the current build project.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild(project_name=u"testproject")
        self.assertEqual(u"testproject",
                         slave_server.api_getBuildProjectName())

    def test_api_get_build_project_name_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildProjectName should return an error
        state if there's no current build project.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildProjectName())

    def test_api_get_build_name(self):
        """
        LexbuilderSlaveServer.getBuildName should return the name of the
        current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild(name=u"testbuild")
        self.assertEqual(u"testbuild",
                         slave_server.api_getBuildName())

    def test_api_get_build_name_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildName should return an error
        state if there's no current build project.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildName())

    def test_api_get_build_result(self):
        """
        LexbuilderSlaveServer.getBuildResult should return the result of the
        current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()
        self.assertEqual(ProjectBuildProcess.STATE_SUCCESS,
                         slave_server.api_getBuildResult())

    def test_api_get_build_result_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildResult should return an error
        state if there's no current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildResult())

    def test_api_get_build_started_at(self):
        """
        LexbuilderSlaveServer.getBuildStartedAt should return the time the
        current build (if any) started at, using the xmlrpc DateTime
        serialiser.
        """
        slave_server = self.create_slave_server()
        current_build = FakeBuild(started_at=datetime.today())
        self.fake_slave.currentBuild = current_build
        self.assertEqual(xmlrpclib.DateTime(current_build.startedAt),
                         slave_server.api_getBuildStartedAt())

    def test_api_get_build_started_at_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildStartedAt should return an error
        state if there's no current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildStartedAt())

    def test_api_get_build_started_at_with_no_start_time(self):
        """
        LexbuilderSlaveServer.api_getBuildStartedAt should return an error if
        the start time of the current build is unknown.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildStartedAt())

    def test_api_get_build_finished_at(self):
        """
        LexbuilderSlaveServer.getBuildFinishedAt should return the time the
        current build (if any) finished at, using the xmlrpc DateTime
        serialiser.
        """
        slave_server = self.create_slave_server()
        current_build = FakeBuild(finished_at=datetime.today())
        self.fake_slave.currentBuild = current_build
        self.assertEqual(xmlrpclib.DateTime(current_build.finishedAt),
                         slave_server.api_getBuildFinishedAt())

    def test_api_get_build_finished_at_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildFinishedAt should return an error
        state if there's no current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildFinishedAt())

    def test_api_get_build_finished_at_with_no_finish_time(self):
        """
        LexbuilderSlaveServer.api_getBuildFinishedAt should return an error if
        the finish time of the current build is unknown.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildFinishedAt())

    def test_api_get_build_error_code(self):
        """
        LexbuilderSlaveServer.getBuildErrorCode should return the current error
        code of the build.
        """
        slave_server = self.create_slave_server()
        current_build = FakeBuild(error_code=1)
        self.fake_slave.currentBuild = current_build
        self.assertEqual(1, slave_server.api_getBuildErrorCode())

    def test_api_get_build_error_code_with_no_active_build(self):
        """
        LexbuilderSlaveServer.api_getBuildErrorCode should return an error
        state if there's no current build.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = None
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildErrorCode())

    def test_api_get_build_error_code_with_no_finish_time(self):
        """
        LexbuilderSlaveServer.api_getBuildErrorCode should return an error if
        the error code of the current build is unknown.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_getBuildErrorCode())

    def test_api_shutdown_slave_not_idle(self):
        """
        LexbuilderSlaveServer.api_shutdown should return an error if the slave
        is not idle.
        """
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()
        self.fake_slave.state = ProjectBuildProcess.STATE_ACTIVE
        self.assertEqual(LexbuilderSlaveServer.REQUEST_ERROR,
                         slave_server.api_shutdown())

    def test_api_shutdown_slave_idle(self):
        """
        LexbuilderSlaveServer.api_shutdown should start a shutdown thread if
        the slave is currently idle.
        """
        class FakeShutdownThread(object):
            def start(self):
                self.started = True
        slave_server = self.create_slave_server()
        self.fake_slave.currentBuild = FakeBuild()

        mock_shutdown_thread = self.mocker.replace(ShutdownThread,
                                                   passthrough=False)
        mock_shutdown_thread(slave_server)
        fake_shutdown = FakeShutdownThread()
        self.mocker.result(fake_shutdown)
        self.mocker.replay()

        self.assertEqual(LexbuilderSlaveServer.REQUEST_OKAY,
                         slave_server.api_shutdown())
        self.assertTrue(fake_shutdown.started)
