# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from django.test import TestCase

from offspring.web.queuemanager.models import LaunchpadProject
from offspring.web.queuemanager.forms import LaunchpadProjectForm


class LaunchpadProjectFormTests(TestCase):

    def test_form_is_invalid_with_no_data(self):
        """
        LaunchpadProjectForm is based on the LaunchpadProject, so requires a
        name value.
        """
        form = LaunchpadProjectForm({})
        self.assertFalse(form.is_valid())
        self.assertEqual({"name": [u"This field is required."]},
                         form.errors)

    def test_form_validates_with_all_fields(self):
        """
        Supplying a valid name to the form makes it valid.
        """
        form = LaunchpadProjectForm({"name": u"Testing"})
        self.assertTrue(form.is_valid())

    def test_form_is_invalid_with_name_too_long(self):
        """
        The name for a LaunchpadProject cannot be longer than 200 characters.
        """
        form = LaunchpadProjectForm({"name": u"x"*201})
        self.assertFalse(form.is_valid())
        self.assertEqual({"name": [u"Ensure this value has at most 200 "
                                    "characters (it has 201)."]},
                         form.errors)

    def test_title_cant_be_edited_from_form(self):
        """
        The title of a LaunchpadProjectForm can't be edited.
        """
        form = LaunchpadProjectForm({"name": u"testing", "title": u"Testing"})
        self.assertTrue(form.is_valid())
        self.assertFalse("title" in form.cleaned_data,
                         "title should not be editable")

    def test_form_cleans_name_to_lower_case(self):
        """
        LaunchpadProjectForm normalises the name supplied to be all lower case.
        """
        form = LaunchpadProjectForm({"name": u"Testing"})
        self.assertTrue(form.is_valid())
        self.assertEqual(u"testing", form.cleaned_data["name"])

    def test_validates_name_is_unique(self):
        """
        LaunchpadProjectForm should check that the name of the LaunchpadProject
        to be added is unique.
        """
        LaunchpadProject.objects.create(name=u"testing")
        form = LaunchpadProjectForm({"name": u"Testing"})
        self.assertFalse(form.is_valid())
        self.assertEqual(
            {"name": [u"Launchpad project with this Launchpad project name "
                       "already exists."]}, form.errors)
