import errno
import os

from django.contrib.auth import authenticate, login
from django.http import HttpRequest, HttpResponse
from django.test import TestCase
from django.contrib.sessions.middleware import SessionMiddleware

from piston.utils import rc

from offspring.web.queuemanager.models import BuildRequest
from offspring.web.queuemanager.handlers import (
    BuildManifestComparisonHandler,
    BuildManifestHandler,
    BuildRequestHandler,
    BuildResultHandler,
    ReleaseHandler,
    )
from offspring.web.queuemanager.tests.factory import factory
from offspring.web.queuemanager.tests.helpers import grant_permission_to_user


class BaseHandlerTestCase(TestCase):
    """
    A TestCase with some custom setup and helpers for testing Piston handlers.
    """

    def setUp(self):
        super(BaseHandlerTestCase, self).setUp()
        self.request = HttpRequest()
        self.request.method = 'GET'
        # Set request.user and process the request manually using
        # SessionMiddleware so that we can use login(request, user) in tests.
        self.request.user = None
        SessionMiddleware().process_request(self.request)

    def _login(self, user):
        user = authenticate(username=user.username, password=user.username)
        login(self.request, user)
        return user


class BuildResultHandlerTests(BaseHandlerTestCase):

    def test_builds_for_inexistent_project(self):
        """
        Check that no builds are returned if there's no project with the given
        name.
        """
        results = BuildResultHandler().read(
            self.request, projectName='not-existent')
        self.assertEqual(0, results.count())

    def test_inexistent_build(self):
        """
        Check that no builds are returned if there's no build with the given
        name.
        """
        results = BuildResultHandler().read(
            self.request, buildName='not-existent')
        self.assertEqual(0, results.count())

    def test_public_project_builds(self):
        """
        Check that the builds of public projects are visible to anyone.
        """
        build = make_build(is_private=False)
        self.assertBuildIsVisibleToUser(build, factory.make_user())

    def test_private_project_build_is_visible_to_project_owner(self):
        """
        Check that the builds of private projects are visible to the project's
        owner.
        """
        build = make_build(is_private=True)
        self.assertBuildIsVisibleToUser(build, build.project.owner)

    def test_private_project_builds_invisible_to_user(self):
        """
        Check that the builds of private projects are not visible to members
        of the project's access group.

        If the user has no rights to see a given project, they'll get an
        empty list of builds, just like if the project didn't exist.
        """
        build = make_build(is_private=True)
        self.assertBuildIsNotVisibleToUser(build, factory.make_user())

    def assertBuildIsVisibleToUser(self, build, user):
        """
        Assert that the given build is visible to the given user.
        """
        user = self._login(user)
        handler = BuildResultHandler()

        # It is visible when looking at all project builds...
        results = handler.read(self.request, projectName=build.project.name)
        self.assertEqual(1, results.count())
        self.assertEqual([build], list(results))

        # when looking at one specific build ...
        results = handler.read(self.request, buildName=build.name)
        self.assertEqual(1, results.count())
        self.assertEqual([build], list(results))

        # and when looking at all builds of a builder.
        results = handler.read(self.request, builderName=build.builder.name)
        self.assertEqual(1, results.count())
        self.assertEqual([build], list(results))

    def assertBuildIsNotVisibleToUser(self, build, user):
        """
        Assert that the given build is not visible to the given user.
        """
        user = self._login(user)
        handler = BuildResultHandler()

        # It is *not* visible when looking at all project builds...
        results = handler.read(self.request, projectName=build.project.name)
        self.assertEqual(0, results.count())

        # when looking at one specific build ...
        results = handler.read(self.request, buildName=build.name)
        self.assertEqual(0, results.count())

        # and when looking at all builds of a builder.
        results = handler.read(self.request, builderName=build.builder.name)
        self.assertEqual(0, results.count())


class BuildRequestHandlerTests(BaseHandlerTestCase):

    def test_public_build_request_is_visible_to_anyone(self):
        """
        Check that build requests of public projects are visible to anyone.
        """
        job = factory.make_build_request(
            project=factory.make_project(is_private=False))
        results = BuildRequestHandler().read(
            self.request, projectName=job.project.name, request_id=job.id)
        self.assertEqual(job, results)

    def test_private_build_request_is_visible_to_owner(self):
        """
        Check that build requests of private projects are visible to the
        project's owner.
        """
        job = factory.make_build_request(
            project=factory.make_project(is_private=True))
        self._login(job.project.owner)
        results = BuildRequestHandler().read(
            self.request, projectName=job.project.name, request_id=job.id)
        self.assertEqual(job, results)

    def test_private_build_request_is_not_visible_to_others(self):
        """
        Check that build requests of private projects are not visible if the
        user doesn't have the rights to see the project.
        """
        job = factory.make_build_request(
            project=factory.make_project(is_private=True))
        self._login(factory.make_user())
        self.assertRaises(
            BuildRequest.DoesNotExist,
            BuildRequestHandler().read,
            self.request, projectName=job.project.name, request_id=job.id)

    def test_new_build_request_on_public_project(self):
        """
        Check that anybody can request a build of a public project.
        """
        project = factory.make_project(is_private=False)
        self.assertBuildHandlerCreation(
            project, factory.make_user(), rc.CREATED.status_code)

    def test_owner_requests_new_build_on_private_project(self):
        """
        Check that the project owner can request the build of a private
        project.
        """
        project = factory.make_project(is_private=True)
        self.assertBuildHandlerCreation(
            project, project.owner, rc.CREATED.status_code)

    def test_other_user_requests_new_build_on_private_project(self):
        """
        Check that users who can't see a project cannot request a build
        of it.
        """
        project = factory.make_project(is_private=True)
        self.assertBuildHandlerCreation(
            project, factory.make_user(), rc.BAD_REQUEST.status_code)

    def test_delete_build_request_on_public_project(self):
        """
        Check that anybody can cancel a build request of a public project.
        """
        job = factory.make_build_request(
            project=factory.make_project(is_private=False))
        self.assertBuildHandlerDeletion(
            job, factory.make_user(), rc.DELETED.status_code)

    def test_owner_deletes_build_request_on_private_project(self):
        """
        Check that the project owner can cancel a build request of a private
        project.
        """
        job = factory.make_build_request(
            project=factory.make_project(is_private=True))
        self.assertBuildHandlerDeletion(
            job, job.project.owner, rc.DELETED.status_code)

    def test_other_user_deletes_build_request_on_private_project(self):
        """
        Check that users who can't see a project cannot cancel a build
        request of it.
        """
        job = factory.make_build_request(
            project=factory.make_project(is_private=True))
        self.assertBuildHandlerDeletion(
            job, factory.make_user(), rc.NOT_FOUND.status_code)

    def assertBuildHandlerCreation(self, project, user, return_code):
        """
        Assert that BuildRequestHandler.create() returns the given code.

        Also grant the given user the 'can_build' permission and logs it in.
        """
        grant_permission_to_user(user, 'can_build')
        self._login(user)
        result = BuildRequestHandler().create(
            self.request, projectName=project.name)
        self.assertEqual(return_code, result.status_code)

    def assertBuildHandlerDeletion(self, job, user, return_code):
        """
        Assert that BuildRequestHandler.delete() returns the given code.

        Also grant the given user the 'can_build' permission and logs it in.
        """
        grant_permission_to_user(user, 'can_build')
        self._login(user)
        result = BuildRequestHandler().delete(
            self.request, projectName=job.project.name, request_id=job.id)
        self.assertEqual(return_code, result.status_code)


class ReleaseHandlerTests(BaseHandlerTestCase):

    def test_public_project_release(self):
        """
        Check that releases of public projects are visible to anyone.
        """
        release = factory.make_release(make_build(is_private=False))
        results = ReleaseHandler().read(
            self.request, projectName=release.build.project.name)
        self.assertEqual(1, results.count())
        self.assertEqual([release], list(results))

    def test_private_project_release_is_visible_to_owner(self):
        """
        Check that releases of private projects are visible to the project's
        owner.
        """
        release = factory.make_release(make_build(is_private=True))
        self._login(release.build.project.owner)
        results = ReleaseHandler().read(
            self.request, projectName=release.build.project.name)
        self.assertEqual(1, results.count())
        self.assertEqual([release], list(results))

    def test_private_project_release_is_not_visible_to_others(self):
        """
        Check that releases of private projects are not visible to users who
        don't have the rights to see the project.
        """
        release = factory.make_release(make_build(is_private=True))
        user = self._login(factory.make_user())
        self.assertFalse(release.is_visible_to(user))
        results = ReleaseHandler().read(
            self.request, projectName=release.build.project.name)
        self.assertEqual(0, results.count())


class BuildManifestHandlerTests(BaseHandlerTestCase):

    def test_public_project_build_manifest(self):
        """
        Check that build manifests of public projects are visible to anyone.
        """
        build = make_build(is_private=False)
        make_build_manifest_file(build)
        results = BuildManifestHandler().read(
            self.request, build.project.name, build.name)
        self.assertEqual([{'package': 'foo', 'version': '1.2'}], results)

    def test_private_project_build_manifest_is_visible_to_owner(self):
        """
        Check that build manifests of private projects are visible to the
        project's owner.
        """
        build = make_build(is_private=True)
        make_build_manifest_file(build)
        self._login(build.project.owner)
        results = BuildManifestHandler().read(
            self.request, build.project.name, build.name)
        self.assertEqual([{'package': 'foo', 'version': '1.2'}], results)

    def test_private_project_build_manifest_returns_404_for_others(self):
        """
        Check that build manifests of private projects are not visible to
        users who don't have the rights to see the project.
        """
        build = make_build(is_private=True)
        make_build_manifest_file(build)
        user = self._login(factory.make_user())
        self.assertFalse(build.project.is_visible_to(user))
        results = BuildManifestHandler().read(
            self.request, build.project.name, build.name)
        self.assertTrue(isinstance(results, HttpResponse))
        self.assertEqual(404, results.status_code)


class BuildManifestComparisonHandlerTests(BaseHandlerTestCase):

    def test_public_projects_build_manifest_comparison(self):
        """
        Check that anyone can get a build comparison between builds of a
        public project.
        """
        build = make_build(is_private=False)
        build2 = factory.make_build_result(
            project=build.project, builder=factory.make_lexbuilder())
        make_build_manifest_file(build)
        make_build_manifest_file(build2)
        self.request.GET['build'] = build2.name
        results = BuildManifestComparisonHandler().read(
            self.request, build.project.name, build.name)
        self.assertEqual(
            [{'changeType': 'no_change', 'target_version': '1.2',
              'base_version': '1.2', 'package': 'foo'}],
            results)

    def test_private_build_manifest_comparison_is_visible_to_owner(self):
        """
        Check that the project owner can get a build comparison between builds
        of a private project.
        """
        build = make_build(is_private=True)
        build2 = factory.make_build_result(
            project=build.project, builder=factory.make_lexbuilder())
        make_build_manifest_file(build)
        make_build_manifest_file(build2)
        self._login(build.project.owner)
        self.request.GET['build'] = build2.name
        results = BuildManifestComparisonHandler().read(
            self.request, build.project.name, build.name)
        self.assertEqual(
            [{'changeType': 'no_change', 'target_version': '1.2',
              'base_version': '1.2', 'package': 'foo'}],
            results)

    def test_private_build_manifest_comparison_returns_404_for_others(self):
        """
        Check that those who don't have the rights to see a private project
        cannot get a build comparison between builds of that project.
        """
        build = make_build(is_private=True)
        build2 = factory.make_build_result(
            project=build.project, builder=factory.make_lexbuilder())
        make_build_manifest_file(build)
        make_build_manifest_file(build2)
        self._login(factory.make_user())
        self.request.GET['build'] = build2.name
        results = BuildManifestComparisonHandler().read(
            self.request, build.project.name, build.name)
        self.assertTrue(isinstance(results, HttpResponse))
        self.assertEqual(404, results.status_code)


def make_build(is_private):
    """Create a new BuildResult associated to a newly created project.

    If is_private is True, the project is marked as private, otherwise it is
    marked as public.
    """
    return factory.make_build_result(
        project=factory.make_project(is_private=is_private),
        builder=factory.make_lexbuilder())


def make_build_manifest_file(build):
    """
    Create a manifest file where BuildManifestHandler expects to find it.
    """
    manifest_dir = os.path.join(build.result_directory, 'images', 'build')
    try:
        os.makedirs(manifest_dir)
    except OSError, exc:
        if exc.errno != errno.EEXIST:
            raise
    path = os.path.join(manifest_dir, 'build.manifest')
    with open(path, 'w') as fd:
        fd.write('foo 1.2')
