from operator import attrgetter

from django.contrib.auth.models import AnonymousUser
from django.test import TestCase

from offspring.web.queuemanager.models import (
    BuildRequest, BuildResult, Lexbuilder, Project, Release)
from offspring.web.queuemanager.tests.factory import factory


class ProjectTests(TestCase):

    def test_default_manager(self):
        """
        Make sure .all_objects is the default manager.

        If that's not the case the admin UI will exclude all private objects
        and we would break model validation of unique fields if the user
        reuses a value from an object they're not allowed to see.
        """
        self.assertEquals(Project.all_objects, Project._default_manager)

    def test_is_private_for_public_project(self):
        """
        Check that Project.is_private returns False for public projects.
        """
        project = factory.make_project(is_private=False)
        self.assertEqual(False, project.is_private)

    def test_is_private_for_private_project(self):
        """
        Check that Project.is_private returns True for private projects.
        """
        private_project = factory.make_project(is_private=True)
        self.assertEqual(True, private_project.is_private)

    def test_is_visible_to_anyone_if_public(self):
        """
        Check that public projects are visible to anyone.
        """
        project = factory.make_project(is_private=False)
        self.assertTrue(project.is_visible_to(AnonymousUser()))

    def test_is_not_visible_to_anyone_if_private(self):
        """
        Check that private projects are not visible to anonymous users.
        """
        project = factory.make_project(is_private=True)
        self.assertFalse(project.is_visible_to(AnonymousUser()))

    def test_is_visible_to_owner_if_private(self):
        """
        Check that private projects are visible to their owners.
        """
        project = factory.make_project(is_private=True)
        self.assertTrue(project.is_visible_to(project.owner))

    def test_is_visible_to_member_of_access_group_if_private(self):
        """
        Check that private projects are visible to members of their access
        groups.
        """
        user = factory.make_user()
        group = factory.make_access_group([user])
        project = factory.make_project(is_private=True, access_groups=[group])
        self.assertTrue(project.is_visible_to(user))

    def test_dot_objects_model_manager_only_returns_public_projects(self):
        """
        Check that the .objects model manager return only public projects.
        """
        project = factory.make_project(is_private=False)
        project2 = factory.make_project(is_private=True)
        self.assertEqual([project], list(Project.objects.all()))

    def test_all_objects_model_manager(self):
        """
        Check that Project.all_objects includes private objects.

        Project.all_objects() is a separate model manager which returns
        private objects as well. Its .all() method will return public and
        private objects regardless of whether or not the current user has
        access to them. It is to be used with caution.
        """
        project = factory.make_project(is_private=False)
        project2 = factory.make_project(is_private=True)
        self.assertEqual([project, project2], list(Project.all_objects.all()))

    def test_accessible_by_user_with_anonymous_user(self):
        """
        Check that accessible_by_user() can be given an anonymous user as
        argument.
        """
        project = factory.make_project(is_private=True)
        project2 = factory.make_project(is_private=False)
        self.assertEqual(True, project.is_private)
        self.assertEqual(False, project2.is_private)
        objects = Project.all_objects.accessible_by_user(AnonymousUser())
        self.assertEqual([project2], list(objects))

    def test_accessible_by_user_with_public_project(self):
        """
        Check that accessible_by_user() returns public projects regardless of
        the given user.
        """
        user = factory.make_user()
        project = factory.make_project(is_private=False)
        self.assertEqual(False, project.is_private)
        objects = Project.all_objects.accessible_by_user(user)
        self.assertEqual([project], list(objects))

    def test_accessible_by_user_with_private_project(self):
        """
        Check that accessible_by_user() returns only the private projects the
        given user is allowed to see.
        """
        user = factory.make_user()
        group = factory.make_access_group([user])
        project = factory.make_project(is_private=True, access_groups=[group])
        # This second project is private but has no access groups set up, so
        # the user cannot see it.
        project2 = factory.make_project(is_private=True)
        self.assertEqual(True, project.is_private)
        self.assertEqual(True, project2.is_private)
        objects = Project.all_objects.accessible_by_user(user)
        self.assertEqual([project], list(objects))


class ReleaseOrLexbuilderOrBuildResultOrBuildRequestTestsMixin(object):
    """A mixin with tests that work for the following models:

        Release
        Lexbuilder
        BuildResult
        BuildRequest

    You just need to mix this with TestCase in your subclass and define
    factoryMethod and model.
    """
    model = None

    def factoryMethod(self, project):
        raise NotImplementedError()

    def test_default_manager(self):
        """
        Make sure .all_objects is the default manager.

        If that's not the case the admin UI will exclude all private objects
        and we would break model validation of unique fields if the user
        reuses a value from an object they're not allowed to see.
        """
        self.assertEquals(self.model.all_objects, self.model._default_manager)

    def test_is_private_for_public_object(self):
        """
        Check that .is_private returns False for public objects.
        """
        public_object = self.factoryMethod(
            factory.make_project(is_private=False))
        self.assertFalse(public_object.is_private)

    def test_is_private_for_private_object(self):
        """
        Check that .is_private returns True for private objects.
        """
        private_object = self.factoryMethod(
            factory.make_project(is_private=True))
        self.assertTrue(private_object.is_private)

    def test_is_visible_to_anyone_if_public(self):
        """
        Check that public objects are visible to anyone.
        """
        public_object = self.factoryMethod(
            factory.make_project(is_private=False))
        self.assertTrue(public_object.is_visible_to(AnonymousUser()))

    def test_is_not_visible_to_anyone_if_private(self):
        """
        Check that private objects are not visible to anonymous users.
        """
        private_object = self.factoryMethod(
            factory.make_project(is_private=True))
        self.assertFalse(private_object.is_visible_to(AnonymousUser()))

    def test_is_visible_to_owner_if_private(self):
        """
        Check that private objects are visible to their project's owners.
        """
        # The object we're dealing with inherits the access-control rules from
        # the project it's related to, so the owner that matters is the
        # project owner.
        owner = factory.make_user()
        private_object = self.factoryMethod(
            factory.make_project(is_private=True, owner=owner))
        self.assertTrue(private_object.is_visible_to(owner))

    def test_is_visible_to_member_of_access_group_if_private(self):
        """
        Check that private objects are visible to members of their project's
        access groups.
        """
        user = factory.make_user()
        group = factory.make_access_group([user])
        private_object = self.factoryMethod(
            factory.make_project(is_private=True, access_groups=[group]))
        self.assertTrue(private_object.is_visible_to(user))

    def test_dot_objects_model_manager_only_returns_public_objects(self):
        """
        Check that the .objects model manager return only public objects.
        """
        private_object = self.factoryMethod(
            factory.make_project(is_private=True))
        public_object = self.factoryMethod(
            factory.make_project(is_private=False))
        self.assertEqual([public_object], list(self.model.objects.all()))

    def test_all_objects_model_manager(self):
        """
        Check that .all_objects includes private objects.

        self.model.all_objects() is a separate model manager which returns
        private objects as well. Its .all() method will return public and
        private objects regardless of whether or not the current user has
        access to them. It is to be used with caution.
        """
        public_object = self.factoryMethod(
            project=factory.make_project(is_private=False))
        private_object = self.factoryMethod(
            project=factory.make_project(is_private=True))
        self.assertEqual([public_object, private_object],
                         list(self.model.all_objects.all()))

    def test_accessible_by_user(self):
        """
        Check that accessible_by_user() returns only the private objects the
        user is allowed to see.
        """
        user = factory.make_user()
        group = factory.make_access_group([user])
        visible_object = self.factoryMethod(
            factory.make_project(is_private=True, access_groups=[group]))
        # This object is linked to a private project which has no access
        # groups set up, so the user cannot see it.
        invisible_object = self.factoryMethod(
            project=factory.make_project(is_private=True))
        self.assertEqual(
            [visible_object],
            list(self.model.all_objects.accessible_by_user(user)))


class ReleaseTests(
    TestCase, ReleaseOrLexbuilderOrBuildResultOrBuildRequestTestsMixin):
    model = Release

    def factoryMethod(self, project):
        return factory.make_release(
            build=factory.make_build_result(project=project))


class LexbuilderTests(
        TestCase, ReleaseOrLexbuilderOrBuildResultOrBuildRequestTestsMixin):
    model = Lexbuilder

    def factoryMethod(self, project):
        return factory.make_lexbuilder(
            current_job=factory.make_build_result(project=project))


class BuildResultTests(
        TestCase, ReleaseOrLexbuilderOrBuildResultOrBuildRequestTestsMixin):
    model = BuildResult
    factoryMethod = factory.make_build_result


class BuildRequestTests(
        TestCase, ReleaseOrLexbuilderOrBuildResultOrBuildRequestTestsMixin):
    model = BuildRequest
    factoryMethod = factory.make_build_request


class ProjectGroupTests(TestCase):

    def test_get_projects_filters_private_objects(self):
        """
        ProjectGroup.get_projects() will return only the private projects the
        given user is allowed to see.
        """
        user = factory.make_user()
        group = factory.make_project_group()
        public_project = factory.make_project(
            is_private=False, project_group=group)
        private_project_visible_to_user = factory.make_project(
            is_private=True, project_group=group, owner=user)
        private_project = factory.make_project(
            is_private=True, project_group=group)
        self.assertEqual(
            sorted([public_project, private_project_visible_to_user],
                   key=attrgetter("name")),
            sorted(group.get_projects(user), key=attrgetter("name")))
