# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import urllib

from django.contrib.auth.models import AnonymousUser
from django.core.urlresolvers import reverse
from django.http import Http404
from django.test import TestCase

from offspring.enums import ProjectBuildStates
from offspring.web.queuemanager.models import (
    Lexbuilder, Project, LaunchpadProject)
from offspring.web.queuemanager.views import get_possibly_private_object
from offspring.web.queuemanager.forms import LaunchpadProjectForm

from offspring.web.queuemanager.tests.factory import factory
from offspring.web.queuemanager.tests.helpers import grant_permission_to_user


class TestGetPossiblyPrivateObject(TestCase):

    def test_public_object(self):
        """
        Check that get_possibly_private_object() returns the public project
        with the given key.
        """
        obj = factory.make_project(is_private=False)
        obj2 = get_possibly_private_object(
            AnonymousUser(), Project, pk=obj.pk)
        self.assertEqual(obj.pk, obj2.pk)

    def test_private_object_for_owner(self):
        """
        Check that get_possibly_private_object() returns the private project
        with the given key when the user is the project's owner.
        """
        obj = factory.make_project(is_private=True)
        obj2 = get_possibly_private_object(obj.owner, Project, pk=obj.pk)
        self.assertEqual(obj.pk, obj2.pk)

    def test_private_object_for_user_with_no_access(self):
        """
        Check that get_possibly_private_object() raises a 404 error when a
        user tries to get a project they're not allowed to see.
        """
        obj = factory.make_project(is_private=True)
        user = factory.make_user()
        self.assertRaises(
            Http404, get_possibly_private_object, user, Project, pk=obj.pk)


class ProjectViewTestsMixin(object):
    view_path = None

    def get_expected_page_heading(self, project):
        """
        Return the heading that should be present on the page tested here.

        Must be overwritten in subclasses.
        """
        raise NotImplementedError()

    def grant_necessary_permission_to(self, user):
        """
        Grant the permission necessary for the given user to see the page
        tested here.

        Only override it in subclasses if the page you're testing requires
        an extra permission to be seen.
        """

    def test_private_project_visible(self):
        """
        Check that this view of a private project is accessible to those who
        have the rights to see the project.
        """
        project = factory.make_project(is_private=True)
        user = project.owner
        self.grant_necessary_permission_to(user)
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(
            reverse(self.view_path, args=[project.name]))
        self.assertContains(
            response, self.get_expected_page_heading(project),
            status_code=200, msg_prefix=response.content)

    def test_private_project_invisible(self):
        """
        Check that this view of a private project is not accessible to those
        who lack the rights to see the project.
        """
        project = factory.make_project(is_private=True)
        user = factory.make_user()
        self.grant_necessary_permission_to(user)
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(
            reverse(self.view_path, args=[project.name]))
        self.assertEqual(404, response.status_code)


class ProjectDetailsViewTests(TestCase, ProjectViewTestsMixin):
    view_path = 'offspring.web.queuemanager.views.project_details'

    def get_expected_page_heading(self, project):
        return 'Project Details for %s' % project.title.capitalize()

    def test_public_project(self):
        """
        Check that the details view of a public project is accessible to
        anyone.
        """
        project = factory.make_project(is_private=False)
        response = self.client.get(
            reverse(self.view_path, args=[project.name]))
        self.assertContains(
            response, self.get_expected_page_heading(project),
            status_code=200, msg_prefix=response.content)


class ProjectEditViewTests(TestCase, ProjectViewTestsMixin):
    view_path = 'offspring.web.queuemanager.views.project_edit'

    def get_expected_page_heading(self, project):
        return 'Update details for %s' % project.title.capitalize()

    def grant_necessary_permission_to(self, user):
        grant_permission_to_user(user, 'change_project')

    def test_public_project(self):
        """
        Check that the edit view of a public project is accessible to those
        who have the change_project permission.
        """
        project = factory.make_project(is_private=False)
        user = factory.make_user()
        self.grant_necessary_permission_to(user)
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(
            reverse(self.view_path, args=[project.name]))
        self.assertContains(
            response, self.get_expected_page_heading(project),
            status_code=200, msg_prefix=response.content)


class BuilderListTests(TestCase):
    view_name = 'builder_list'

    def get_url(self, builder):
        # The builder argument exists only so that the tests here also work
        # for the builder_details view (see BuilderDetailsTests below).
        return reverse(self.view_name)

    def setUp(self):
        super(BuilderListTests, self).setUp()
        # Stub Lexbuilder.average_load because it executes some SQL that's not
        # supported by sqlite3 and we're not interested in testing it here.
        self.orig_load_average = Lexbuilder.average_load
        def average_load(obj):
            return '10'
        Lexbuilder.average_load = property(average_load)

    def tearDown(self):
        super(BuilderListTests, self).tearDown()
        Lexbuilder.average_load = self.orig_load_average

    def test_build_details_always_shown_when_building_public_project(self):
        """
        Check that details of public project builds are always shown.
        """
        builder = factory.make_lexbuilder()
        make_user_and_login(self.client)
        response = self.client.get(self.get_url(builder))
        self.assertContains(
            response, builder.current_job.project.name, status_code=200,
            msg_prefix=response.content)

    def test_private_build_details_shown_if_user_is_allowed_to_see_it(self):
        """
        Check that details of private project builds are shown when the user
        is allowed to see that project.
        """
        project = factory.make_project(is_private=True)
        job = factory.make_build_result(project=project)
        builder = factory.make_lexbuilder(current_job=job)
        user = project.owner
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(self.get_url(builder))
        self.assertContains(
            response, builder.current_job.project.name, status_code=200,
            msg_prefix=response.content)

    def test_private_build_details_not_shown_if_user_not_allowed_to_see(self):
        """
        Check that details of private project builds are not shown if the user
        is not allowed to see that project.

        Notice that the builders themselves are shown; we just redact the
        details of the job.
        """
        project = factory.make_project(is_private=True)
        job = factory.make_build_result(project=project)
        builder = factory.make_lexbuilder(current_job=job)
        make_user_and_login(self.client)
        response = self.client.get(self.get_url(builder))
        self.assertNotContains(
            response, builder.current_job.project.name, status_code=200,
            msg_prefix=response.content)


class BuilderDetailsTests(BuilderListTests):
    view_name = 'builder_details'

    def get_url(self, builder):
        return reverse(self.view_name, args=[builder.name])


class BuildResultViewTests(TestCase):

    def test_public_build(self):
        """
        Check that the main view of a public build is accessible to anyone.
        """
        project = factory.make_project(is_private=False)
        build = factory.make_build_result(project=project)
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.build_details',
                    args=[project.name, build.name]))
        self.assertContains(
            response, 'Build Details', status_code=200,
            msg_prefix=response.content)

    def test_private_build_visible(self):
        """
        Check that the main view of a private build is accessible to those who
        have the rights to see the build's project.
        """
        project = factory.make_project(is_private=True)
        build = factory.make_build_result(project=project)
        user = project.owner
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.build_details',
                    args=[project.name, build.name]))
        self.assertContains(
            response, 'Build Details', status_code=200,
            msg_prefix=response.content)

    def test_private_build_invisible(self):
        """
        Check that the main view of a private build is not accessible to those
        who lack the rights to see the build's project.
        """
        project = factory.make_project(is_private=True)
        build = factory.make_build_result(project=project)
        self.assertTrue(make_user_and_login(self.client))
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.build_details',
                    args=[project.name, build.name]))
        self.assertEqual(404, response.status_code)


class BuildResultPublishViewTests(TestCase):

    def test_public_build(self):
        """
        Check that the publish view of a public build is accessible to anyone.
        """
        project = factory.make_project(is_private=False)
        build = factory.make_build_result(
            project=project, result=ProjectBuildStates.SUCCESS)
        user = factory.make_user()
        grant_permission_to_user(user, 'add_release')
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.build_publish',
                    args=[project.name, build.name]))
        self.assertContains(
            response, 'Release %s build' % project.title.capitalize(),
            status_code=200, msg_prefix=response.content)

    def test_private_build_visible(self):
        """
        Check that the publish view of a private build is accessible to those
        who have the rights to see the build's project and the add_release
        permission.
        """
        project = factory.make_project(is_private=True)
        build = factory.make_build_result(
            project=project, result=ProjectBuildStates.SUCCESS)
        user = build.project.owner
        grant_permission_to_user(user, 'add_release')
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.build_publish',
                    args=[project.name, build.name]))
        self.assertContains(
            response, 'Release %s build' % project.title.capitalize(),
            status_code=200, msg_prefix=response.content)

    def test_private_build_invisible(self):
        """
        Check that the publish view of a private build is not accessible to
        those who lack the rights to see the build's project.
        """
        project = factory.make_project(is_private=True)
        build = factory.make_build_result(
            project=project, result=ProjectBuildStates.SUCCESS)
        user = factory.make_user()
        grant_permission_to_user(user, 'add_release')
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.build_publish',
                    args=[project.name, build.name]))
        self.assertEqual(404, response.status_code)


class TestProjectACLView(TestCase):

    def test_is_visible_to_owner(self):
        """
        Check that the +acl view is visible to the project's owner.
        """
        project = factory.make_project(is_private=True)
        user = project.owner
        grant_permission_to_user(user, 'change_project')
        self.assertTrue(
            self.client.login(username=user.username, password=user.username),
            "Failed to login")
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.project_acl',
                    args=[project.name]))
        self.assertContains(
            response, 'Update ACLs for %s' % project.title.capitalize(),
            status_code=200, msg_prefix=response.content)

    def test_is_visible_to_access_group_member(self):
        """
        Check that the +acl view is visible to those who are members of the
        project's access groups.
        """
        user = factory.make_user()
        grant_permission_to_user(user, 'change_project')
        group = factory.make_access_group([user])
        project = factory.make_project(is_private=True, access_groups=[group])
        self.assertTrue(
            self.client.login(username=user.username, password=user.username),
            "Failed to login")
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.project_acl',
                    args=[project.name]))
        self.assertContains(
            response, 'Update ACLs for %s' % project.title.capitalize(),
            status_code=200, msg_prefix=response.content)

    def test_is_not_visible_to_others(self):
        """
        Check that the +acl view is not visible to those who are not the
        project owner or a member of its access group, even if they have the
        change_project permission.
        """
        user = factory.make_user()
        grant_permission_to_user(user, 'change_project')
        project = factory.make_project(is_private=True)
        self.assertTrue(
            self.client.login(username=user.username, password=user.username),
            "Failed to login")
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.project_acl',
                    args=[project.name]))
        self.assertEqual(404, response.status_code)

    def test_add_new_user(self):
        """
        Check that adding a new user to a project's ACL adds that user to the
        project's access group.
        """
        project = factory.make_project(is_private=True)
        user = project.owner
        grant_permission_to_user(user, 'change_project')
        self.assertTrue(
            self.client.login(username=user.username, password=user.username),
            "Failed to login")
        member = factory.make_user()
        data = {'new_user': member.id}
        response = self.client.post(
            reverse('offspring.web.queuemanager.views.project_acl',
                    args=[project.name]),
            data)
        self.assertEqual(200, response.status_code)
        access_group = project.access_groups.all()[0]
        self.assertEqual([member], list(access_group.members.all()))

    def test_remove_member(self):
        """
        Check that removing a user from the project's ACL removes it from the
        project's access group.
        """
        member = factory.make_user()
        group = factory.make_access_group([member])
        project = factory.make_project(is_private=True, access_groups=[group])
        user = project.owner
        grant_permission_to_user(user, 'change_project')
        self.assertTrue(
            self.client.login(username=user.username, password=user.username),
            "Failed to login")
        data = {'disallow': 'Disallow', member.username: 'on'}
        response = self.client.post(
            reverse('offspring.web.queuemanager.views.project_acl',
                    args=[project.name]),
            data)
        self.assertEqual(200, response.status_code)
        access_group = project.access_groups.all()[0]
        self.assertEqual([], list(access_group.members.all()))


class ProjectBuildsViewTests(TestCase, ProjectViewTestsMixin):
    view_path = 'offspring.web.queuemanager.views.builds'

    def get_expected_page_heading(self, project):
        return 'Build Records for %s' % project.title

    def test_public_project(self):
        """
        Check that the builds page of public projects is visible to anyone.
        """
        project = factory.make_project(is_private=False)
        factory.make_build_result(project=project)
        response = self.client.get(
            reverse(self.view_path, args=[project.name]))
        self.assertContains(
            response, self.get_expected_page_heading(project),
            status_code=200, msg_prefix=response.content)


class OverallBuildsViewTests(TestCase):
    view_path = 'offspring.web.queuemanager.views.builds'

    def test_public_builds_are_shown(self):
        """
        Check that the overall builds page always shows the public ones.
        """
        project = factory.make_project(is_private=False)
        build = factory.make_build_result(project=project)
        response = self.client.get(reverse(self.view_path))
        self.assertContains(
            response, build.name, status_code=200,
            msg_prefix=response.content)

    def test_private_builds_are_shown_if_user_has_rights(self):
        """
        Check that the overall builds page shows the private ones the user is
        allowed to see.
        """
        project = factory.make_project(is_private=True)
        build = factory.make_build_result(project=project)
        user = project.owner
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(reverse(self.view_path))
        self.assertContains(
            response, build.name, status_code=200,
            msg_prefix=response.content)

    def test_private_builds_are_not_shown_if_user_has_no_rights(self):
        """
        Check that the overall builds page does not show the private ones the
        user is not allowed to see.
        """
        project = factory.make_project(is_private=True)
        build = factory.make_build_result(project=project)
        make_user_and_login(self.client)
        response = self.client.get(reverse(self.view_path))
        self.assertNotContains(
            response, build.name, status_code=200,
            msg_prefix=response.content)


class ProjectListTests(TestCase):
    view_path = 'offspring.web.queuemanager.views.projects'

    def test_public_projects_are_shown_to_anyone(self):
        """
        Check that public projects are always shown on the project list.
        """
        project = factory.make_project(is_private=False)
        response = self.client.get(reverse(self.view_path))
        self.assertContains(
            response, project.name, status_code=200,
            msg_prefix=response.content)

    def test_private_projects_are_shown_if_user_has_rights(self):
        """
        Check that private projects are only shown on the project list when
        the user has the rights to see them.
        """
        project = factory.make_project(is_private=True)
        user = project.owner
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(reverse(self.view_path))
        self.assertContains(
            response, project.name, status_code=200,
            msg_prefix=response.content)

    def test_private_projects_are_not_shown_if_user_has_no_rights(self):
        """
        Check that private projects which the user is not allowed to see are
        not shown on the project list.
        """
        project = factory.make_project(is_private=True)
        make_user_and_login(self.client)
        response = self.client.get(reverse(self.view_path))
        self.assertNotContains(
            response, project.name, status_code=200,
            msg_prefix=response.content)


def make_user_and_login(client):
    user = factory.make_user()
    return client.login(username=user.username, password=user.username)


class LaunchpadProjectViewsTests(TestCase):

    def setUp(self):
        self.url = reverse("launchpad_project_create")
        self.user = factory.make_user()

    def test_launchpad_project_non_superuser_gets_redirected(self):
        """
        If the logged in user doesn't have permission to access the form, they
        should be directed to login.
        """
        self.client.login(username=self.user.username,
                          password=self.user.username)
        response = self.client.get(self.url)

        self.assertEqual(302, response.status_code)
        self.assertEqual(
            "http://testserver/admin/?next=" + urllib.quote(self.url),
            response["Location"])

    def test_get_launchpad_project_add_url_gets_correct_form(self):
        """
        If the logged in user has the correct permissions, a GET request
        should get the LaunchpadProjectForm.
        """
        grant_permission_to_user(self.user, "add_launchpadproject")
        self.client.login(username=self.user.username,
                          password=self.user.username)

        response = self.client.get(self.url)
        self.assertEqual(200, response.status_code)
        self.assertTrue(
            isinstance(response.context["form"], LaunchpadProjectForm),
            "Form is not LaunchpadProjectForm")
        self.assertEqual("queuemanager/launchpad_project_create_popup.html",
                         response.template.name)

    def test_post_launchpad_project_add_url(self):
        """
        A POST request should create a new LaunchpadProject with the entered
        name.
        """
        grant_permission_to_user(self.user, "add_launchpadproject")
        self.client.login(username=self.user.username,
                          password=self.user.username)

        self.assertEqual([], list(LaunchpadProject.objects.all()))
        response = self.client.post(self.url, {"name": u"My New Project"})
        project = LaunchpadProject.objects.all()[0]
        self.assertEqual(u"my new project", project.name)
        self.assertTrue('"my new project"' in response.content,
                        "Project name didn't appear in the response.")
