# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import urlparse

from django.contrib.auth.decorators import (
    login_required,
    permission_required
)
from django.core.paginator import (
    Paginator,
    InvalidPage,
    EmptyPage
)
from django.db.models import Q
from django.http import (
    HttpResponse,
    HttpResponseRedirect,
    HttpResponseForbidden,
    Http404,
)
from django.middleware import csrf
from django.shortcuts import render_to_response, get_object_or_404
from django.template import (
    Context,
    RequestContext,
    Template
)
from django.utils.html import escape
from django.views.generic.create_update import (
    create_object,
    update_object
)

from pygooglechart import PieChart3D

from django_group_access.models import AccessGroup

from offspring.config import get_configuration
from offspring.enums import ProjectBuildStates
from offspring.web.queuemanager.forms import (
    AccessGroupMemberForm,
    CreateProjectForm,
    EditProjectForm,
    LaunchpadProjectForm,
    ReleaseForm,
)
from offspring.web.queuemanager.models import (
    BuildRequest,
    BuildResult,
    LaunchpadProjectMilestone,
    Lexbuilder,
    Project,
    ProjectGroup,
    Release
)


config = get_configuration()


def get_possibly_private_object(user, klass, *args, **kwargs):
    """Use get_object_or_404() to return a possibly private object.

    If the object does not exist or the user doesn't have permission to see
    it, raise Http404.
    """
    obj = get_object_or_404(klass, *args, **kwargs)
    if not obj.is_visible_to(user):
        raise Http404('No matches for the given query')
    return obj


#XXX: Reduce horrible duplication that occurs in these functions (mostly done).
#TODO: Thanks to reduced duplication, lets start using more generic views.

@login_required
def builders(request):
    # builder_list is a dictionary where keys are Lexbuilder instances to be
    # shown and the values represent whether or not the information about the
    # builder's current job should be shown to the current user.
    builder_list = {}
    # Here we don't filter out the builders working on private projects (as we
    # do with other stuff, using .accessible_by_user()) because we want to
    # show that the builders exist and just hide the info about private
    # projects the user is not allowed to see.
    builders = Lexbuilder.all_objects.order_by(
        "-is_active", "-machine_type", "-created_at")
    for builder in builders:
        job = builder.current_job
        if job is not None and job.is_visible_to(request.user):
            builder_list[builder] = True
        else:
            builder_list[builder] = False
    pageData = {
        'pillar': 'builders',
        'builder_list': builder_list,
        }
    return render_to_response(
        'queuemanager/builders.html', pageData,
        context_instance=RequestContext(request))

@login_required
def builder_details(request, builderName):
    # Here we don't filter out the builders working on private projects (as we
    # do with other stuff, using .accessible_by_user()) because we want to
    # show that the builders exist and just hide the info about private
    # projects the user is not allowed to see.
    builder = Lexbuilder.all_objects.get(name=builderName)
    shown_job = 'Private build'
    job = builder.current_job
    if job is not None and job.is_visible_to(request.user):
        shown_job = job
    pageData = {
        'builder': builder,
        'current_job': shown_job,
        'pillar': 'builders',
        }
    return render_to_response(
        'queuemanager/builder_details.html', pageData,
        context_instance=RequestContext(request))

def projects(request):
    projects = Project.all_objects.accessible_by_user(request.user)
    projects = projects.select_related('project_group').order_by(
        "project_group")
    if request.path == "/" or request.path == "":
        pillar = "home"
        #XXX: Move logic for determining active projects into model manager.
        projects = projects.exclude(Q(is_active=False) & Q(dailybuildorder=None))
    else:
        pillar = "projects"
    pageData = { 
        'projects' : projects,
        'pillar' : pillar
    }
    return render_to_response(
        'queuemanager/projects.html', pageData,
        context_instance=RequestContext(request))

def builds(request, projectName=None):
    user = request.user
    user_visible_builds = BuildResult.all_objects.accessible_by_user(user)
    if projectName is not None:
        project = get_possibly_private_object(user, Project, pk=projectName)
        buildResults_list = user_visible_builds.filter(
            project=projectName).order_by("finished_at").reverse()
    else:
        project = None
        buildResults_list = user_visible_builds.select_related(
            'builder', 'requestor', 'project').order_by("finished_at").reverse()
    paginator = Paginator(buildResults_list, 15)

    try:
        page = int(request.GET.get('page', '1'))
    except ValueError:
        page = 1

    try:
        buildResults = paginator.page(page)
    except (EmptyPage, InvalidPage):
        buildResults = paginator.page(paginator.num_pages)

    is_private = False
    if project is not None and project.is_private:
        is_private = True
    pageData = {
        'project': project,
        'is_private': is_private,
        'build_results': buildResults,
        'pillar': 'builds'
    }
    return render_to_response(
        'queuemanager/builds.html', pageData,
        context_instance=RequestContext(request))


def build_details(request, projectName, buildName):
    user_visible_builds = BuildResult.all_objects.accessible_by_user(
        request.user)
    build = get_possibly_private_object(
        request.user,
        user_visible_builds.select_related('builder', 'project', 'requestor'),
        project__pk=projectName, name=buildName)
    pageData = { 
        'pillar' : 'builds',
        'upcoming_milestones' : LaunchpadProjectMilestone.upcoming_milestones(),
        'recent_milestones' : LaunchpadProjectMilestone.recent_milestones(),
        'project' : build.project, 
        'is_private': build.is_private,
        'build' : build
    }
    return render_to_response(
        'queuemanager/build_details.html', pageData, 
        context_instance=RequestContext(request))

@login_required
def build_publish(request, projectName, buildName):
    user = request.user
    p = get_possibly_private_object(user, Project, pk=projectName)
    # We use get_possibly_private_object() here just for consistency as it's
    # not really needed because the above will fail if the user doesn't have
    # the rights to see this build.
    b = get_possibly_private_object(
        user, BuildResult, project=p, name=buildName, 
        result=ProjectBuildStates.SUCCESS, release=None)
    if user.has_perm('queuemanager.add_release'):
        r = Release()
        r.build = b
        r.creator = user
        if request.method == 'POST':
            f = ReleaseForm(request.POST, instance=r)
            if f.is_valid():
                f.save()
                return HttpResponseRedirect(p.get_absolute_url())
        else:
            f = ReleaseForm(instance=r)
        f.fields['milestone'].queryset = LaunchpadProjectMilestone.objects.filter( \
            launchpad_project=f.instance.build.project.launchpad_project \
            ).exclude(active=False)
        pageData = { 
            'pillar' : 'builds',
            'project' : p,
            'build' : b,
            'form' : f 
        }
        return render_to_response(
            'queuemanager/build_publish.html', pageData, 
            context_instance=RequestContext(request))
    else:
        return HttpResponseForbidden(render_to_response('403.html'))

def project_create(request):
    return create_object(
        request,
        form_class=CreateProjectForm, 
        login_required=True,
        template_name='queuemanager/project_create.html',
        extra_context={ 'pillar' : 'projects' },
    )
project_create = permission_required('queuemanager.add_project')(project_create)

def project_edit(request, projectName):
    # XXX: This nasty hack is because update_object uses (via lookup_object)
    # .objects directly when it should insted use the default manager
    # (https://code.djangoproject.com/ticket/17021).
    from django.views.generic import create_update
    orig_lookup = create_update.lookup_object
    def lookup(model, obj_id, slug, slug_field):
        return get_possibly_private_object(
            request.user, Project, pk=projectName)
    create_update.lookup_object = lookup
    try:
        response = update_object(
            request,
            form_class=EditProjectForm,
            login_required=True,
            template_name='queuemanager/project_edit.html',
            template_object_name='project',
            extra_context={ 'pillar' : 'projects' },
            object_id=projectName,
        )
    finally:
        create_update.lookup_object = orig_lookup
    return response
project_edit = permission_required('queuemanager.change_project')(project_edit)

def project_details(request, projectName):
    user_visible_objects = Project.all_objects.accessible_by_user(
        request.user)
    p = get_possibly_private_object(
        request.user, user_visible_objects, pk=projectName)
    user_visible_builds = BuildResult.all_objects.accessible_by_user(
        request.user)
    project_build_results = user_visible_builds.filter(
        project = p).exclude(result = None)
    build_stats = { 
        'fail_count' : project_build_results.filter(result = ProjectBuildStates.FAILED).count(),
        'success_count' : project_build_results.filter(result = ProjectBuildStates.SUCCESS).count() 
    }

    if len(project_build_results):
        chart = PieChart3D(425, 100)
        chart.add_data([build_stats['fail_count'], build_stats['success_count']])
        chart.set_colours(['D01B1B', '60B33C'])
        chart.set_pie_labels(["Failed Builds", "Successful Builds"])
        build_stats['pie_chart'] = chart.get_url()

    releases_uri_base = config.get('web', 'releases_uri_base')
    if not releases_uri_base.endswith('/'):
        releases_uri_base += '/'
    releases_url = urlparse.urljoin(releases_uri_base, p.name + '/images/')

    archive_uri = config.get('web', 'archive_uri', None)
    pageData = {
        'build_stats' : build_stats,
        'csrf_token' : csrf.get_token(request),
        'include_sources_list': archive_uri is not None,
        'pillar' : 'projects',
        'project' : p,
        'is_private': p.is_private,
        'project_build_results' : project_build_results,
        'releases_url': releases_url,
        }
    return render_to_response(
        'queuemanager/project_details.html', pageData, 
        context_instance=RequestContext(request))


@permission_required("queuemanager.change_project")
def project_acl(request, projectName):
    """A page to edit a project's access control list."""
    user_visible_objects = Project.all_objects.accessible_by_user(
        request.user)
    project = get_possibly_private_object(
        request.user, user_visible_objects, pk=projectName)
    access_group = None
    allowed_users = []
    if project.access_groups.count() > 0:
        # We always use the first access group because there's no way for
        # users to register more than one access group for any given project.
        access_group = project.access_groups.all()[0]
        allowed_users = access_group.members.all()
    if request.method == "POST":
        form = AccessGroupMemberForm(request.POST)
        if form.is_valid():
            # A form submission can be either to add a new member or to remove
            # an existing one.
            if form.data.get("new_user"):
                if access_group is None:
                    # This project has no access group set up yet, so we
                    # create one.
                    access_group = AccessGroup(name=project.name)
                    access_group.save()
                    project.access_groups.add(access_group)
                    project.save()
                access_group.members.add(form.cleaned_data["new_user"])
                access_group.save()
            else:
                assert "disallow" in form.data.keys(), "Unknown form action"
                for user in allowed_users:
                    if user.username in form.data.keys():
                        access_group.members.remove(user)
                        access_group.save()
            # We don't redirect on a successful form submission because most
            # of the time users will want to make multiple changes to the ACL.
            allowed_users = access_group.members.all()
            form = AccessGroupMemberForm()
    else:
        form = AccessGroupMemberForm()
    pageData = {
        "csrf_token" : csrf.get_token(request),
        "project" : project,
        "form": form,
        "allowed_users": allowed_users,
        }
    return render_to_response(
        "queuemanager/project_acl.html", pageData,
        context_instance=RequestContext(request))


def projectgroup_details(request, projectGroupName):
    pg = get_object_or_404(
        ProjectGroup, pk=projectGroupName)
    projectgroup_build_results = BuildResult.objects.filter(
        project__project_group = pg).exclude(result = None)
    build_stats = { 
        'fail_count' : projectgroup_build_results.filter(result = ProjectBuildStates.FAILED).count(),
        'success_count' : projectgroup_build_results.filter(result = ProjectBuildStates.SUCCESS).count()
    }

    if len(projectgroup_build_results):
        chart = PieChart3D(425, 150)
        chart.add_data([build_stats['fail_count'], build_stats['success_count']])
        chart.set_colours(['D01B1B', '60B33C'])
        chart.set_pie_labels(["Failed Builds", "Successful Builds"])
        build_stats['pie_chart'] = chart.get_url()

    pageData = { 
        'projectGroup' : pg, 
        'projects': pg.get_projects(request.user),
        'pillar' : 'projects', 
        'build_stats' : build_stats,
        'projectgroup_build_results' : projectgroup_build_results,
        'pillar' : 'projects'
    }
    return render_to_response(
        'queuemanager/projectgroup_details.html', pageData, 
        context_instance=RequestContext(request))

@login_required
def project_development_sourcesList(request, projectName):
    c = Context({"projectName": projectName})
    if request.GET.get('action') == "view":
        t = Template(
            "{% load lexbuilder_helpers %}<link rel=\"stylesheet\" type=\"text/css\" href=\"/assets/css/project_details.css\"/><div id=\"sourcesListModalContent\"><pre>{{ projectName|generate_devel_sourceslist }}</pre></div>")
        return HttpResponse(t.render(c))
    else:
        t = Template(
            "{% load lexbuilder_helpers %}{{ projectName|generate_devel_sourceslist }}")
        return HttpResponse(
            t.render(c), content_type="text/x-apt-sources-list")

@login_required
def queue_build(request, projectName):
    p = get_object_or_404(Project, pk=projectName)
    if p.needs_build:
        BuildRequest.objects.filter(project = p).delete()
    else:
        if request.POST.has_key('build_reason'):
            build_reason = request.POST['build_reason']
        else:
            build_reason = ""
        request = BuildRequest.queue_build(
            p, request.user, build_reason, scoreBonus=25)
        request.save()
    return HttpResponseRedirect(p.get_absolute_url())
queue_build = permission_required('queuemanager.can_build')(queue_build)

def schedule(request):
    pageData = { 
        'pillar' : 'schedule',
        'upcoming_milestones' : LaunchpadProjectMilestone.upcoming_milestones(),
        'recent_milestones' : LaunchpadProjectMilestone.recent_milestones()
    }
    return render_to_response(
        'queuemanager/schedule.html', pageData,
        context_instance=RequestContext(request))

def milestones(request, milestoneType="all"):
    if milestoneType == "future":
        milestones = LaunchpadProjectMilestone.upcoming_milestones(limit=None)
    elif milestoneType == "past":
        milestones = LaunchpadProjectMilestone.recent_milestones(limit=None)
    else:
        milestoneType = "all"
        milestones =  LaunchpadProjectMilestone.objects.all().order_by("date_targeted").reverse()
    paginator = Paginator(milestones, 10)

    try:
        page = int(request.GET.get('page', '1'))
    except ValueError:
        page = 1

    try:
        milestonePage = paginator.page(page)
    except (EmptyPage, InvalidPage):
        milestonePage = paginator.page(paginator.num_pages)

    pageData = {
        'milestones' : milestonePage,
        'milestoneType' : milestoneType,
        'pillar' : 'schedule'
    }
    return render_to_response(
        'queuemanager/milestones.html', pageData,
        context_instance=RequestContext(request))

def launchpad_project_create(request):
    """
    Create a new LaunchpadProject.
    """
    if request.method == "POST":
        form = LaunchpadProjectForm(request.POST)
        if form.is_valid():
            new_project = form.save()
            javascript = ('<script type="text/javascript">opener.dismissAddAno'
                          'therPopup(window, "%s", "%s");</script>')
            return HttpResponse(javascript % (
                escape(new_project._get_pk_val()), escape(new_project)))
    else:
        form = LaunchpadProjectForm()

    page_context = {"form": form}
    return render_to_response(
        "queuemanager/launchpad_project_create_popup.html",
        page_context,
        context_instance=RequestContext(request))
launchpad_project_create = permission_required(
    "queuemanager.add_launchpadproject")(launchpad_project_create)
